
#include "libvdev/sglib.h"
#include "libvdev/util.h"
#include "libvdev/misc.h"
#include "libvdev/sbuf.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <errno.h>
#include <unistd.h>
#include <poll.h>
#include <sys/signalfd.h>

#include <libudev.h>

const char *progname = "vdev_libinput_wait";

//#define UDEV_XKB_PROP_KEY "xkb"

/* Size of buffer to use when reading inotify events */
#define INOTIFY_BUFFER_SIZE 8192

/* Enumerate list of FDs to poll */
enum {
  
   FD_POLL_SIGNAL = 0,
   FD_POLL_UDEV,
   FD_POLL_MAX
};

static
int initialize_signals (void)
{
   int signal_fd;
   sigset_t sigmask;

   /* We want to handle SIGINT and SIGTERM in the signal_fd, so we block them. */
   sigemptyset (&sigmask);
   sigaddset (&sigmask, SIGINT);
   sigaddset (&sigmask, SIGTERM);

   if (sigprocmask (SIG_BLOCK, &sigmask, NULL) < 0) {
      
      vdev_error ("%s: Couldn't block signals: '%s'\n", progname, strerror (errno));
      return -1;
   }

   /* Get new FD to read signals from it */
   if ((signal_fd = signalfd (-1, &sigmask, 0)) < 0) {
      
      vdev_error ("%s: Couldn't setup signal FD: '%s'\n", progname, strerror (errno));
      return -1;
   }

   return signal_fd;
}

#define BLOCK_SIZE 512

typedef char* cstr;

// prototypes
SGLIB_DEFINE_VECTOR_PROTOTYPES (cstr);
SGLIB_DEFINE_VECTOR_FUNCTIONS (cstr);

int main(int argc, char **argv)
{
   FILE *fp = NULL;
   char line[1024];
   struct sglib_cstr_vector input_devices;
   int num_input_devices = 0;
  
   struct udev *udev = NULL;
   struct udev_enumerate *enumerate = NULL;
   struct udev_list_entry *devices, *dev_list_entry;
   
   char pathbuf[PATH_MAX+1];
   struct udev_monitor* monitor = NULL;
   struct pollfd pfd[FD_POLL_MAX];  
   int signal_fd, monitor_fd;
   
   alarm (10);
   
   fp = fopen("/proc/bus/input/devices", "r");
   if (!fp) {

      vdev_error("%s:fopen(): '%s'\n", progname, strerror (errno));
      return 1;
   }
 
   sglib_cstr_vector_init(&input_devices);
   
   while (fgets(line, 1024, fp) != NULL) { 
	   
	  line[strcspn(line, "\n")] = 0;
      if (!strncmp(line, "N: Name=", strlen("N: Name="))) { 
	
         char *name = rindex(line, '=') + 1;
         if (strstr(name, "Keyboard") || strstr(name, "Mouse") || strstr(name, "Touchpad")) {
			 			 
			 sglib_cstr_vector_push_back (&input_devices, vdev_strdup_or_null(name));
		 }	 
      }
   }       

   fclose(fp);

   /* create udev object */
   udev = udev_new ();
   if (!udev) {

      vdev_error("%s: Cannot create udev context: '%s'\n", progname, strerror (errno));
      
      sglib_cstr_vector_free(&input_devices);
      return 2;
   }

   /* create enumerate object */
   enumerate = udev_enumerate_new (udev);
   if (!enumerate) {

      vdev_error ("%s: Cannot create enumerate context: '%s'\n", progname, strerror (errno));
      
      sglib_cstr_vector_free(&input_devices);
      return 3;
   }

   udev_enumerate_add_match_subsystem (enumerate, "input");
   udev_enumerate_scan_devices (enumerate);

   /* fillup device list */
   devices = udev_enumerate_get_list_entry (enumerate);
   if (!devices) {
 
      vdev_error ("%s: Failed to get device list: '%s'\n", progname, strerror (errno));
      
      sglib_cstr_vector_free(&input_devices);
      udev_enumerate_unref (enumerate);
      udev_unref (udev);   
      return 4;
   }

   udev_list_entry_foreach (dev_list_entry, devices) {

      const char *path, *tmp;
      unsigned long long disk_size = 0;
      unsigned short int block_size = BLOCK_SIZE;
      struct udev_device *device = NULL;
      const char *tags_prop;
      struct udev_list_entry *set, *entry;
      const char *key, *value, *aux;

      path = udev_list_entry_get_name (dev_list_entry);
      device = udev_device_new_from_syspath (udev, path);
      if (device == NULL)
          continue;
    
      set = udev_device_get_properties_list_entry(device);
      udev_list_entry_foreach(entry, set) {
		  
          key = udev_list_entry_get_name(entry);
          if (!key)
              continue;
              
          value = udev_list_entry_get_value(entry);
          if (!strcmp(key, "NAME")) {

			 for (int i = 0; i < sglib_cstr_vector_size(&input_devices); i++ ) { 
			    
			    if (!strcmp(value, sglib_cstr_vector_at(&input_devices, i))) {

                   num_input_devices++;
                   printf("%s\n", value);
                   break;
                }   
             }      
             
             if (num_input_devices == sglib_cstr_vector_size(&input_devices)) {
      
                 sglib_cstr_vector_free(&input_devices);
                 				 
				 udev_device_unref (device);
                 udev_enumerate_unref (enumerate);
                 udev_unref (udev);   
                 return 0;
             }
          }  
		/*			  
			  printf("%s=%s\n", key, value);
          if (!strncasecmp(key, UDEV_XKB_PROP_KEY, sizeof(UDEV_XKB_PROP_KEY) - 1)) {
              aux = key + sizeof(UDEV_XKB_PROP_KEY) - 1;
              if (!strcasecmp(aux, "rules"))
                  {}
              else if (!strcasecmp(aux, "layout"))
                  {}
              else if (!strcasecmp(aux, "variant"))
                  {}
              else if (!strcasecmp(aux, "model"))
                  {}
              else if (!strcasecmp(aux, "options"))
                  {}
          }
          */
      }

      udev_device_unref (device);
   }

   udev_enumerate_unref (enumerate);
  
   /* NETLINK */
    
   udev_monitor_fs_events_path( "", pathbuf, 0 );
   
   monitor = udev_monitor_new_from_netlink ( udev, "udev" );
   if( monitor == NULL ) {
 
      // OOM or error
      vdev_error ("%s: Couldn't initialize udev monitor: '%s'\n", progname, strerror(errno));
      
      sglib_cstr_vector_free(&input_devices);
      udev_unref (udev);
      exit(EXIT_FAILURE);
   }
   
   monitor_fd = udev_monitor_get_fd (monitor);
   if( monitor_fd < 0 ) {
 
      vdev_error ("%s: udev_monitor_get_fd: '%s'\n", progname, strerror(errno));
      
      sglib_cstr_vector_free(&input_devices);
      udev_unref (udev);
      exit (EXIT_FAILURE);
   }

   /* Initialize signals FD */
   if ((signal_fd = initialize_signals ()) < 0) {
     
      vdev_error ("%s: Couldn't initialize signals: '%s'\n", progname, strerror(errno));
      
      sglib_cstr_vector_free(&input_devices);
      udev_monitor_fs_destroy (monitor);
      udev_unref (udev);
      exit (EXIT_FAILURE);
   }
    
   /* Setup polling */
   pfd[FD_POLL_SIGNAL].fd = signal_fd;
   pfd[FD_POLL_SIGNAL].events = POLLIN;
   
   pfd[FD_POLL_UDEV].fd = monitor_fd;
   pfd[FD_POLL_UDEV].events = POLLIN;

   while(1) {
 
      // next event
      /* Block until there is something to be read */
      if (poll (pfd, FD_POLL_MAX, -1) < 0) {

          vdev_error ("%s: Couldn't poll(): '%s'\n", progname, strerror (errno));
          goto Exit;
      }

      /* Signal received? */
      if (pfd[FD_POLL_SIGNAL].revents & POLLIN) {
         
         struct signalfd_siginfo fdsi;

         if (read (pfd[FD_POLL_SIGNAL].fd, &fdsi, sizeof (fdsi)) != sizeof (fdsi)) {
 
              vdev_error ("%s: Couldn't read signal, wrong size read: '%s'\n", progname, strerror (errno));
              goto Exit;
         }

         /* Break loop if we got the expected signal */
         if (fdsi.ssi_signo == SIGINT  ||
             fdsi.ssi_signo == SIGTERM ||
             fdsi.ssi_signo == SIGALRM) {

                printf("\n");
                break;
         }

         vdev_info ("%s: Received unexpected signal\n", progname);
      }  
      
      // get devices 
      while(1) {
   
         struct udev_device *dev = NULL;
         struct udev_list_entry *set, *entry;
         const char *key, *value, *aux;
         
         dev = udev_monitor_receive_device (monitor);
         if (dev == NULL)
            break;

         /* skip if device/disk is a partition or loop device */
         if (!strcmp(udev_device_get_subsystem( dev ), "input")) {
			 
		    set = udev_device_get_properties_list_entry(dev);
            udev_list_entry_foreach(entry, set) {
         
               key = udev_list_entry_get_name(entry);
               if (!key)
                  continue;
            
               value = udev_list_entry_get_value(entry);
          
               if (!strcmp(key, "NAME")) {
			      
			      for (int i = 0; i < sglib_cstr_vector_size(&input_devices); i++ ) { 
			    
			         if (!strcmp(value, sglib_cstr_vector_at(&input_devices, i))) {

                        num_input_devices++;
                        printf("%s\n", value);
                        break;
                     }   
                  }      
             
                  if (num_input_devices == sglib_cstr_vector_size(&input_devices))
                     goto Exit;
               }      
            }
         }
         
         udev_device_unref( dev );
      }   
   }
 
Exit:
      
   sglib_cstr_vector_free(&input_devices);
      
   // we do not close pfd[FD_POLL_UDEV] because it's closed afterwards by udev_monitor_fs_destroy()
   // Hence, we exit the loop when i<FP_POLL_MAX - 1
   for (unsigned i = 0; i < FD_POLL_MAX - 1; i++) {

      if( pfd[i].fd >= 0 ) {
          
         close( pfd[i].fd );
         pfd[i].fd = -1;
      }
   }
   

   udev_monitor_fs_destroy (monitor);

   udev_unref (udev);
   
   usleep(500000);

   return 0;
}
